use std::net::SocketAddr;
use std::sync::Arc;

use anyhow::Result;
use prost::Message;
use serde::Deserialize;
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use tokio::net::{TcpListener, TcpStream};
use tokio::{select, pin, spawn};
use tokio::sync::mpsc::{channel, Receiver, Sender};
use tokio::sync::RwLock;
use tokio::task::JoinHandle;
use tracing::{info, trace, warn};
use etcd_client::{Client as EtcdClient, ResignOptions};

use inner_api::proto;
use crate::prelude::Str;
use crate::table::{AnonTable};
use crate::terminator::Terminator;

#[derive(Deserialize, Debug, Clone)]
pub struct ClusteringConfig {
    /// The name of the cluster used for prefixing the data in etcd and to prevent accidental communication between different clusters.
    ///
    /// Useful for multi-tenant usage of etcd.
    pub cluster_name: Str,
    /// The unique identifier of the node in the cluster, needed to distinguish it form other nodes.
    pub self_id: u32,
    /// The local address where this node listens incoming for node-to-node communication.
    pub listen_on: SocketAddr,
    /// This address is advertised via a gossip mechanism and used by other nodes in the cluster to communicate with this node.
    ///
    /// This is useful when a node is behind a NAT or a proxy, where the publicly accessible IP
    /// address can differ from the local IP address.
    pub advertised_address: SocketAddr,
    /// The address of the etcd node which will be used to discover or to bootstrap the cluster.
    pub etcd_address: Str,
}

/// Cluster registry is a service which handles all connections to other nodes of the cluster.
///
/// It maintains the local view of the cluster state, including the set of nodes,
/// the status of each node and the shard allocation metadata.
pub struct ClusterRegistry {
    inner: Arc<RwLock<ClusterRegistryInner>>,
    listener_task: Terminator,
}

struct ClusterRegistryInner {
    peer_tasks: AnonTable<(Sender<PeerMsg>, JoinHandle<()>)>,
}

impl ClusterRegistry {
    /// Creates a new cluster registry. It contains all currently active peer connections.
    ///
    /// Starts a fiber which listens on the given TCP socket for incoming connections, handles incoming handshake and starts a peer task.
    pub async fn launch(config: ClusteringConfig) -> Result<ClusterRegistry> {
        let inner = Arc::new(RwLock::new(ClusterRegistryInner {
            peer_tasks: AnonTable::new(),
        }));
        let listener = TcpListener::bind(config.listen_on).await?;
        let listener_task = start_connection_listener(listener, inner.clone()).await;
        let mut client = EtcdClient::connect(&[config.etcd_address], None).await?;
        dbg!(client.maintenance_client().status().await?);
        dbg!(client.leader(b"kek").await);
        dbg!(client.kv_client().get(b"kek", None).await?);
        let a = dbg!(client.election_client().campaign(b"kek", b"my name", 0).await?);
        dbg!(client.leader(b"kek").await?);
        dbg!(client.election_client().resign(Some(ResignOptions::new().with_leader(a.leader().unwrap().clone()))).await);


        Ok(ClusterRegistry {
            inner,
            listener_task,
        })
    }

    pub async fn terminate(self) -> Result<()> {
        if let Err(e) = self.listener_task.terminate().await {
            warn!("Failed to terminate clustering listener: {e:?}");
        }
        let mut inner = self.inner.write().await;
        for (_, (k, j)) in inner.peer_tasks.into_iter() {
            if let Err(e) = k.send(PeerMsg::Close).await {
                warn!("Failed to stop peer actor");
            }
        }
        Ok(())
    }
}

async fn start_connection_listener(listener: TcpListener, cluster: Arc<RwLock<ClusterRegistryInner>>) -> Terminator {
    Terminator::spawn(|terminator| async move {
        pin!(terminator);
        loop {
            select! {
                _ = &mut terminator => break,
                res = listener.accept() => {
                    let (mut stream, addr) = match res {
                        Ok(stream) => stream,
                        Err(e) => {
                            warn!("Error accepting connection: {}", e);
                            continue;
                        }
                    };
                    info!("Incoming connection from {addr}");
                    // We handle the handshaking in order to not block other incoming connections.
                    let cluster = cluster.clone();
                    spawn(async move {
                        // All incoming connections expect the handshake to be initiated by the other peer.
                        // The peer sends a `HandshakeRequest` message and we respond to it with a `HandshakeResponse` message.
                        // After that a new task is launched for peer communication, which is the same fiber as the one started when we initiate the connection.
                        let handshake = read_message::<proto::HandshakeRequest>(&mut stream).await;
                        let handshake = match handshake {
                            Ok(handshake) => handshake,
                            Err(e) => {
                                warn!("Error reading from socket: {}", e);
                                return;
                            }
                        };
                        let handshake_response = proto::HandshakeResponse {};
                        let mut buf = vec![];
                        handshake_response.encode(&mut buf).unwrap();
                        stream.flush().await;
                        if let Err(e) = stream.write_all(&buf).await {
                            warn!("Failed to reply with HandshakeResponse: {}", e);
                        } else {
                            info!("Connection from {addr} accepted");
                            peer_launch(stream , cluster).await;
                        }
                    });
                }
            }
        }
        Ok(())
    })
}

async fn read_message<T: Message + Default>(stream: &mut TcpStream) -> Result<T> {
    // All messages received from the stream start with 2 bytes which indicate N -- the length of the message in bytes.
    let mut length_buf = [0; 2];
    if let Err(e) = stream.read_exact(&mut length_buf).await {
        warn!("Error reading length from stream: {}", e);
        return Err(e.into());
    }

    let length = u16::from_be_bytes(length_buf) as usize;

    // After that we expect a raw message of length N.
    let mut msg_buf = vec![0; length];
    if let Err(e) = stream.read_exact(&mut msg_buf).await {
        warn!("Error reading message from stream: {e:?}");
        return Err(e.into());
    }

    match <T as Message>::decode(&*msg_buf) {
        Ok(message) => {
            Ok(message)
        }
        Err(e) => {
            warn!("Invalid message: {e:?}");
            Err(e.into())
        }
    }
}

async fn peer_launch(stream: TcpStream, cluster: Arc<RwLock<ClusterRegistryInner>>) {
    let (tx, rx) = channel(32);
    let join_handle = spawn(peer_communication_loop(stream, rx));
    let mut inner = cluster.write().await;
    inner.peer_tasks.insert((tx, join_handle));
}

enum PeerMsg {
    Close,
}

/// The main loop which is executed from the peer task. Created from both the incoming and the outgoing socket handlers.
async fn peer_communication_loop(mut stream: TcpStream, rx: Receiver<PeerMsg>) {
    loop {
        match read_message::<proto::Request>(&mut stream).await {
            Ok(req) => {
                trace!("Message received: {req:?}");
            }
            Err(e) => {
                warn!("Failed when receiving the next message, closing the socket: {e:?}");
                break;
            }
        }
    }
}
